=== modified file 'bzrlib/graph.py'
--- bzrlib/graph.py	2008-05-21 19:11:28 +0000
+++ bzrlib/graph.py	2008-05-21 19:41:45 +0000
@@ -227,15 +227,37 @@
         NULL_REVISION = revision.NULL_REVISION
         known_revnos[NULL_REVISION] = 0
 
+        searching_known_tips = list(known_revnos.keys())
+
+        unknown_searched = {}
+
         while cur_tip not in known_revnos:
+            unknown_searched[cur_tip] = num_steps
+            num_steps += 1
             to_search = set([cur_tip])
+            to_search.update(searching_known_tips)
             parent_map = self.get_parent_map(to_search)
             parents = parent_map.get(cur_tip, None)
             if not parents: # An empty list, or None is still a ghost
                 raise errors.GhostRevisionsHaveNoRevno(target_revision_id,
                                                        cur_tip)
-            num_steps += 1
             cur_tip = parents[0]
+            next_known_tips = []
+            for revision_id in searching_known_tips:
+                parents = parent_map.get(revision_id, None)
+                if not parents:
+                    continue
+                next = parents[0]
+                next_revno = known_revnos[revision_id] - 1
+                if next in unknown_searched:
+                    # We have enough information to return a value right now
+                    return next_revno + unknown_searched[next]
+                if next in known_revnos:
+                    assert known_revnos[next] == next_revno
+                    continue
+                known_revnos[next] = next_revno
+                next_known_tips.append(next)
+            searching_known_tips = next_known_tips
 
         # We reached a known revision, so just add in how many steps it took to
         # get there.

=== modified file 'bzrlib/tests/test_graph.py'
--- bzrlib/tests/test_graph.py	2008-05-21 19:11:28 +0000
+++ bzrlib/tests/test_graph.py	2008-05-21 19:41:45 +0000
@@ -396,6 +396,28 @@
 with_ghost = {'a': ['b'], 'c': ['b', 'd'], 'b':['e'], 'd':['e', 'g'],
               'e': ['f'], 'f':[NULL_REVISION], NULL_REVISION:()}
 
+# A graph that shows we can shortcut finding revnos when reaching them from the
+# side.
+#  NULL_REVISION
+#       |
+#       a
+#       |
+#       b
+#       |
+#       c
+#       |
+#       d
+#       |
+#       e
+#      / \
+#     f   g
+#     |
+#     h
+#     |
+#     i
+
+with_tail = {'a':[NULL_REVISION], 'b':['a'], 'c':['b'], 'd':['c'], 'e':['d'],
+             'f':['e'], 'g':['e'], 'h':['f'], 'i':['h']}
 
 
 class InstrumentedParentsProvider(object):
@@ -1268,9 +1290,27 @@
         self.assertFindRevno(3, graph, 'rev3', [('rev2a', 2)])
 
     def test_known_in_ancestry_limits(self):
-        graph = self.make_breaking_graph(ancestry_1, 'rev1')
+        graph = self.make_breaking_graph(ancestry_1, ['rev1'])
         self.assertFindRevno(4, graph, 'rev4', [('rev3', 3)])
 
+    def test_target_is_ancestor(self):
+        graph = self.make_graph(ancestry_1)
+        self.assertFindRevno(2, graph, 'rev2a', [('rev3', 3)])
+
+    def test_target_is_ancestor_limits(self):
+        """We shouldn't search all history if we run into ourselves"""
+        graph = self.make_breaking_graph(ancestry_1, ['rev1'])
+        self.assertFindRevno(3, graph, 'rev3', [('rev4', 4)])
+
+    def test_target_parallel_to_known_limits(self):
+        # Even though the known revision isn't part of the other ancestry, the
+        # eventually converge
+        graph = self.make_breaking_graph(with_tail, ['a'])
+        self.assertFindRevno(6, graph, 'f', [('g', 6)])
+        self.assertFindRevno(7, graph, 'h', [('g', 6)])
+        self.assertFindRevno(8, graph, 'i', [('g', 6)])
+        self.assertFindRevno(6, graph, 'g', [('i', 8)])
+
 
 class TestCachingParentsProvider(tests.TestCase):
 

